from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.contrib.auth.models import User

import datetime

PACKAGE_GREEN_THRESHOLD = 5
PACKAGE_RED_THRESHOLD = 20

EXPERIENCE_CHOICES = (
    (0, ""),
    (1, _("Easy")),
    (2, _("Medium")),
    (3, _("Hard")),
)

class PackageSet(models.Model):
    name = models.SlugField(_("Name"), max_length=40)

    def __unicode__(self):
        return self.name


class SourcePackage(models.Model):
    name = models.SlugField(_("Name"), max_length=70)
    packagesets = models.ManyToManyField(PackageSet, null=True)

    class Meta:
        ordering = ['name']

    def __unicode__(self):
        return self.name


class OpportunityList(models.Model):
    name = models.SlugField(_("Name"), max_length=70)
    url = models.URLField(_("URL"), verify_exists=False)
    description = models.TextField(_("Description"), max_length=350)
    last_updated = models.DateTimeField(_("Last Updated"), null=True)
    active = models.BooleanField(_("Active"), default=True)
    featured = models.BooleanField(_("Featured"), default=False, 
                                   help_text=_("Specially feature this list of opportunities?"))
    explanation = models.URLField(_("Explanatory instructions"), max_length=500, null=True, 
                                   help_text=_("Link to more instructions about this type of opportunities."),
                                   verify_exists=False)
    commentable = models.BooleanField(_("Commentable"), default=True, 
                                      help_text=_("Can opportunities on this list have comments?"))
    experience = models.IntegerField(_("Required Experience"), default=0, choices=EXPERIENCE_CHOICES,
                                     help_text=_("Level of experience required for this type of opportunities."))

    class Meta:
        ordering = ['name']

    def __unicode__(self):
        return self.name


class Opportunity(models.Model):
    description = models.TextField(_("Description"), max_length=350)
    url = models.URLField(_("URL"), max_length=350, verify_exists=False)
    last_updated = models.DateTimeField(_("Last Updated"), null=True)
    since = models.DateTimeField(_("Since"), help_text=_("On the list since"), null=True)
    reviewed = models.BooleanField(_("Irrelevant"), default=False, blank=True)
    applied = models.BooleanField(_("Applied"), default=False, blank=True)
    sourcepackage = models.ForeignKey(SourcePackage)
    opportunitylist = models.ForeignKey(OpportunityList)
    valid = models.BooleanField(_("Valid"), default=True)
    experience = models.IntegerField(_("Difficulty"), choices=EXPERIENCE_CHOICES, default=0,
                                     help_text=_("Level of experience required for this specific opportunity."))

    class Meta:
        ordering = ['-last_updated']
        verbose_name_plural = 'opportunities'

    def __unicode__(self):
        return self.description
    
    @property
    def summary(self):
        summary_list = []
        if self.experience > 0:
            summary_list.append(EXPERIENCE_CHOICES[self.experience][1])
        if self.reviewed:
            summary_list.append(_("Irrelevant"))
        if self.applied:
            summary_list.append(_("Applied"))
        if not self.valid:
            summary_list.append(_("Invalid"))
        return summary_list


class Note(models.Model):
    opportunity = models.ForeignKey(Opportunity)
    date = models.DateTimeField(auto_now_add=True)
    author = models.ForeignKey(User)
    text = models.CharField(max_length=250)
    
    def __unicode__(self):
        text = self.text
        if len(text) > 40:
            text = text[:40]+u"\u2026"
        return '%s: %s' % (self.author, text)


class ActionLogEntry(models.Model):
    timestamp = models.DateTimeField(_("Timestamp"))
    who = models.ForeignKey(User)
    action = models.TextField(_("Action"), max_length=200)
    opportunity = models.ForeignKey(Opportunity, blank=True)

def log_action(who, action=None, opportunity=None):
    log_entry = ActionLogEntry(timestamp=datetime.datetime.now(),
                               who=who, action=action,
                               opportunity=opportunity)
    log_entry.save()

